# frozen_string_literal: true

module Repositories
  class HgHttpController < Repositories::GitHttpClientController
    include WorkhorseRequest
    before_action :access_check

    rescue_from Gitlab::GitAccess::ForbiddenError, with: :render_403_with_exception
    rescue_from Gitlab::GitAccess::NotFoundError, with: :render_404_with_exception
    rescue_from Gitlab::HgAccess::ForbiddenError, with: :render_403_with_exception
    rescue_from Gitlab::HgAccess::NotFoundError, with: :render_404_with_exception

    # TODO 12.10 adopt these exceptions in HgAccess as well
    rescue_from Gitlab::GitAccessProject::CreationError, with: :render_422_with_exception
    rescue_from Gitlab::GitAccess::TimeoutError, with: :render_503_with_exception

    # GET /foo/bar?cmd=XXX
    def hg
      render_ok
    end

    def hg_post
      render_ok
    end

    private

    def download_request?
      Gitlab::HgAccess::DOWNLOAD_COMMANDS.include?(params[:cmd]) && params[:action] == 'hg'
    end

    def hg_command
      if %w[hg hg_post].include?(action_name)
        params[:cmd]
      else
        action_name.dasherize
      end
    end

    def render_ok
      set_workhorse_internal_api_content_type
      render json: Gitlab::Workhorse.git_http_ok(repository, repo_type, user, action_name, hg_access: access)
    end

    def render_403_with_exception(exception)
      render plain: exception.message, status: :forbidden
    end

    def render_404_with_exception(exception)
      render plain: exception.message, status: :not_found
    end

    def render_422_with_exception(exception)
      render plain: exception.message, status: :unprocessable_entity
    end

    def render_503_with_exception(exception)
      render plain: exception.message, status: :service_unavailable
    end

    def update_fetch_statistics
      return unless project
      return if Gitlab::Database.read_only?
      return unless repo_type.project?

      if Feature.enabled?(:project_statistics_sync, project, default_enabled: true)
        Projects::FetchStatisticsIncrementService.new(project).execute
      else
        ProjectDailyStatisticsWorker.perform_async(project.id) # rubocop:disable CodeReuse/Worker
      end
    end

    def access
      @access ||= access_klass.new(access_actor, container, 'http',
        authentication_abilities: authentication_abilities,
        repository_path: repository_path,
        redirected_path: redirected_path,
        auth_result_type: auth_result_type)
    end

    def access_actor
      return user if user
      return :ci if ci?
    end

    def access_check
      access.check(hg_command, '_any')
      @project ||= access.project
    end

    def access_klass
      @access_klass ||= repo_type.hg_access_checker_class
    end

    def repository_path
      # TODO maybe change the route definition to be as generic as the
      # current Git one, with only the `repository_path` param (would be a
      # step towards more general Mercurial repos, not tied to Project or Wiki)
      @repository_path ||= [params[:namespace_id],
                            params[:repository_id].sub(/\.hg$/, '')].join("/")
    end

    def log_user_activity
      Users::ActivityService.new(user).execute
    end
  end
end
