# frozen_string_literal: true

module Import
  class ValidateRemoteGitEndpointService
    # Validates if the remote endpoint is a valid GIT repository
    # Only smart protocol is supported
    # Validation rules are taken from https://git-scm.com/docs/http-protocol#_smart_clients

    GIT_SERVICE_NAME = "git-upload-pack"
    GIT_EXPECTED_FIRST_PACKET_LINE = "# service=#{GIT_SERVICE_NAME}"
    GIT_BODY_MESSAGE_REGEXP = /^[0-9a-f]{4}#{GIT_EXPECTED_FIRST_PACKET_LINE}/.freeze
    # https://github.com/git/git/blob/master/Documentation/technical/protocol-common.txt#L56-L59
    GIT_PROTOCOL_PKT_LEN = 4
    GIT_MINIMUM_RESPONSE_LENGTH = GIT_PROTOCOL_PKT_LEN + GIT_EXPECTED_FIRST_PACKET_LINE.length
    EXPECTED_CONTENT_TYPE = "application/x-#{GIT_SERVICE_NAME}-advertisement"
    MERCURIAL_CONTENT_TYPE_REGEXP = %r{application/mercurial-\d+.\d+}.freeze
    EXPECTED_MERCURIAL_CAPABILITIES = %w[known getbundle].freeze

    def initialize(params)
      @params = params
      @mercurial = Gitlab::Vcs::MERCURIAL_VCS_TYPES.include?(params.fetch(:vcs_type, 'git'))
    end

    # rubocop:disable Metrics/PerceivedComplexity
    # as usual in Heptapod, refactoring could be a conflicts hazard, we'll see later
    def execute
      uri = Gitlab::Utils.parse_url(@params[:url])

      return ServiceResponse.error(message: "#{@params[:url]} is not a valid URL") unless uri

      return ServiceResponse.success if uri.scheme == 'git'

      uri.fragment = nil
      sub_path_qs = @mercurial ? "?cmd=capabilities" : "/info/refs?service=#{GIT_SERVICE_NAME}"
      url = Gitlab::Utils.append_path(uri.to_s, sub_path_qs)

      response_body = ''
      result = nil
      Gitlab::HTTP.try_get(url, stream_body: true, follow_redirects: false, basic_auth: auth) do |fragment|
        response_body += fragment

        if @mercurial
          # We can read the full body of the capabilities request, it is
          # meant to be short.
          # Of course proper Mercurial always provide Content-Length.
          # Defaulting to 0 in case there is a bogus server. In that case
          # we'll proceed to the check that will surely fail.
          next if response_body.length < fragment.http_response.fetch('content-length', "0").to_i
        elsif response_body.length < GIT_MINIMUM_RESPONSE_LENGTH
          next
        end

        result = if status_code_is_valid(fragment) && content_type_is_valid(fragment) && response_body_is_valid(response_body)
                   :success
                 else
                   :error
                 end

        # In the Git case,
        # We are interested only in the first chunks of the response
        # So we're using stream_body: true and breaking when receive enough body
        break unless @mercurial
      end

      if result == :success
        ServiceResponse.success
      else
        vcs_label = @mercurial ? "Mercurial" : "Git"
        ServiceResponse.error(message: "#{uri} is not a valid HTTP #{vcs_label} repository")
      end
    end
    # rubocop:enable Metrics/PerceivedComplexity

    private

    def auth
      unless @params[:user].to_s.blank?
        {
          username: @params[:user],
          password: @params[:password]
        }
      end
    end

    def status_code_is_valid(fragment)
      fragment.http_response.code == '200'
    end

    def content_type_is_valid(fragment)
      content_type = fragment.http_response['content-type']
      if @mercurial
        !MERCURIAL_CONTENT_TYPE_REGEXP.match(content_type).nil?
      else
        content_type == EXPECTED_CONTENT_TYPE
      end
    end

    def response_body_is_valid(response_body)
      if @mercurial
        capas = response_body.split
        EXPECTED_MERCURIAL_CAPABILITIES.all? { |exp| capas.include?(exp) }
      else
        response_body.match?(GIT_BODY_MESSAGE_REGEXP)
      end
    end
  end
end
