# frozen_string_literal: true

class Repositories::DestroyService < Repositories::BaseService
  def execute
    return success unless repository
    return success unless repo_exists?(disk_path)

    # Flush the cache for both repositories. This has to be done _before_
    # removing the physical repositories as some expiration code depends on
    # Git data (e.g. a list of branch names).
    ignore_git_errors { repository.before_delete }

    # Use variables that aren't methods on Project, because they are used in a callback
    # (Heptapod-specific) upstream then stores storage and path, and initialize
    # a raw repo from the callback. Still the raw repo instance doesn't need
    # the Project (or anything from the database), so we are free to use
    # a full raw repository, thus avoiding duplicating the dispatching logic
    # of Repository.init_raw_repository
    raw_repo = repository.raw_repository

    # Because #remove happens inside a run_after_commit callback it will
    # never be triggered on a read-only instance.
    #
    # Issue: https://gitlab.com/gitlab-org/gitlab/-/issues/223272
    if Gitlab::Database.read_only?
      raw_repo.remove
    else
      container.run_after_commit do
        raw_repo.remove
      end
    end

    log_info("Repository \"#{full_path}\" was removed")

    success
  rescue Gitlab::Git::Repository::NoRepository
    success
  end
end
