# frozen_string_literal: true

class PostReceive
  include ApplicationWorker

  idempotent!
  deduplicate :none
  data_consistency :always

  sidekiq_options retry: 3
  include Gitlab::Experiment::Dsl

  feature_category :source_code_management
  urgency :high
  worker_resource_boundary :cpu
  weight 5
  loggable_arguments 0, 1, 2, 3

  def perform(gl_repository, identifier, hg_prunes, changes, push_options = {})
    container, project, repo_type = Gitlab::GlRepository.parse(gl_repository)

    if container.nil? || (container.is_a?(ProjectSnippet) && project.nil?)
      log("Triggered hook for non-existing gl_repository \"#{gl_repository}\"")
      return false
    end

    changes = Base64.decode64(changes) unless changes.include?(' ')
    # Use Sidekiq.logger so arguments can be correlated with execution
    # time and thread ID's.
    Sidekiq.logger.info "changes: #{changes.inspect}" if SidekiqLogArguments.enabled?
    post_received = Gitlab::GitPostReceive.new(container, identifier, changes, push_options)

    if repo_type.project? && project.mercurial? && !hg_prunes.nil?
      hg_process_project_prunes(container, post_received, hg_prunes)
    end

    if repo_type.wiki?
      process_wiki_changes(post_received, container)
    elsif repo_type.project?
      process_project_changes(post_received, container)
    elsif repo_type.snippet?
      process_snippet_changes(post_received, container)
    else
      # Other repos don't have hooks for now
    end
  end

  private

  def identify_user(post_received)
    post_received.identify.tap do |user|
      log("Triggered hook for non-existing user \"#{post_received.identifier}\"") unless user
    end
  end

  # rubocop: disable CodeReuse/ActiveRecord
  # again a case where two successive conditionals appear and the "guard clause" cop would
  # want us to introduce an early return for the second
  # TODO split this method in two cases (published_topics, closed branches)
  # rubocop: disable Style/GuardClause
  def hg_process_project_prunes(project, post_received, prunes)
    user = identify_user(post_received)

    logged_method = "PostReceive#hg_process_project_prunes"
    project_id = project.id

    prunes = Gitlab::Json.parse(prunes)

    published_topics = prunes["topic_published"]
    unless published_topics.nil?
      published_topics.each_pair do |gl_branch, latest_sha|
        gl_branch = base64_to_utf8(gl_branch)
        Gitlab::AppLogger.info(
          method: logged_method,
          project_id: project_id,
          message: "looking for Merge Requests from published topic",
          gitlab_branch: gl_branch
        )
        hg_branch, _topic = Gitlab::Mercurial.parse_gitlab_branch_for_hg(gl_branch)
        # locate merge request with topic as source.
        # The odds are that generally, a MR for a published topic will target the
        # corresponding named branch. Therefore is most cases, we won't need to
        # preload the MR diffs like GitLab's standard detection does: we'll reach the
        # conclusion without them. Only if the MR is merged will we need the diff,
        # for potential last update to `latest_sha`.

        # TODO OPTIM make just one SQL request for all published topics
        # and then, sort/match them (could be defeated by the way ActiveRecord
        # fetches from the DB, though).
        # MRs being merged from the Rails app are *not* considered, as
        # they are in the 'locked' state.
        merged = project.merge_requests.opened
          .where(source_branch: gl_branch, source_project: project).to_a
          .select do |merge_request|
          if merge_request.target_branch == "branch/#{hg_branch}"
            Gitlab::AppLogger.info(
              method: logged_method,
              project_id: project_id,
              message: "marking Merge Request as merged: "\
                       "source is the published GitLab branch and "\
                       "target is the corresponding named branch",
              merge_request_iid: merge_request.iid,
              gitlab_branch: gl_branch
            )
            true
          else
            real_target_sha = merge_request.target_branch_sha
            # TODO what if previous position is merged and subsequent
            # changes are not (or maybe in another branch?)
            merged = project.repository.ancestor?(latest_sha, real_target_sha)
            Gitlab::AppLogger.info(
              method: logged_method,
              project_id: project_id,
              message: "Analysis of Merge Request whose source is the "\
                       "published GitLab branch. Latest SHA is `latest_sha`"\
                       "and target is the GitLab branch for another "\
                       "named branch, whose current head SHA is "\
                       "`real_target_sha`",
              merged: merged,
              target_branch: merge_request.target_branch,
              latest_sha: latest_sha,
              real_target_sha: real_target_sha,
              merge_request_iid: merge_request.iid,
              gitlab_branch: gl_branch
            )

            merged
            # If merged is false, we don't need to close `merge_request`:
            # MergeRequests::RefreshService will take care of that later
            # with all necessary details.
          end
        end

        merged.each do |merge_request|
          MergeRequests::HgManuallyMergedService.new(project: project, current_user: user).execute(merge_request,
                                                                                                   latest_sha)
        end
      end
    end

    closed_branches = prunes['branch_closed']
    unless closed_branches.nil?
      merged_with_sha = []
      closed_branches.each_pair do |gl_branch, info|
        next if info.nil?

        gl_branch = base64_to_utf8(gl_branch)

        info.each do |closing_sha, parent_shas|
          merged_with_sha.concat(merged_for_closing_changeset(
                                   project, gl_branch,
                                   closing_sha, parent_shas))
        end
      end
      merged_with_sha.each do |merge_request, latest_sha|
        MergeRequests::HgManuallyMergedService.new(project: project, current_user: user).execute(merge_request,
                                                                                                 latest_sha)
      end
    end
  end
  # rubocop: enable CodeReuse/ActiveRecord
  # rubocop: enable Style/GuardClause

  # Given a closing commit SHA and the SHAs of its parents, find the
  # Merge Requests that have to be updated and merged.
  # rubocop: disable CodeReuse/ActiveRecord
  def merged_for_closing_changeset(project, gl_branch, closing_sha, parent_shas)
    logged_method = "PostReceive#merged_for_closing_changeset"
    project_id = project.id
    repo = project.repository
    merged = []
    project.merge_requests.opened
      .where(source_branch: gl_branch, source_project: project)
      .each do |merge_request|
      real_target_sha = merge_request.target_branch_sha
      if repo.ancestor?(closing_sha, real_target_sha)
        merged << [merge_request, closing_sha]
        Gitlab::AppLogger.info(
          method: logged_method,
          project_id: project_id,
          message: "Marking as merged a Merge Request whose target "\
                   "GitLab branch head is now a direct descendent of the "\
                   "closing changeset",
          merge_request_iid: merge_request.iid,
          target_branch: merge_request.target_branch,
          target_branch_head_sha: real_target_sha,
          closing_sha: closing_sha,
          closed_gitlab_branch: gl_branch
        )
      elsif parent_shas.all? { |psha| repo.ancestor?(psha, real_target_sha) }
        Gitlab::AppLogger.info(
          method: logged_method,
          project_id: project_id,
          message: "Marking as merged a Merge Request whose target "\
                   "GitLab branch head is now a descendent of all the "\
                   "parents of the closing changeset",
          merge_request_iid: merge_request.iid,
          target_branch: merge_request.target_branch,
          target_branch_head_sha: real_target_sha,
          closing_sha: closing_sha,
          closing_parent_shas: parent_shas,
          closed_gitlab_branch: gl_branch
        )
        # if the closing commit has several parents, since we have to update
        # the MR with a single head, let's put in the closing commit to see
        # all of them and be faithful to what happened.
        merged << [merge_request,
                   parent_shas.length == 1 ? parent_shas[0] : closing_sha]
      end
    end
    merged
  end
  # rubocop: enable CodeReuse/ActiveRecord

  # Deserialize a binary string from Base64, then guess and convert to UTF-8
  #
  # This is the same as what happens for the `changes` param, here (base64)
  # and in `Gitlab::GitPostReceive` (transcoding).
  def base64_to_utf8(str)
    decoded = Base64.decode64(str)
    decoded.force_encoding('UTF-8')
    return decoded if decoded.valid_encoding?

    detection = CharlockHolmes::EncodingDetector.detect(changes)
    return decoded unless detection && detection[:encoding]

    CharlockHolmes::Converter.convert(decoded, detection[:encoding], 'UTF-8')
  end

  # rubocop: disable CodeReuse/ActiveRecord
  def hg_fixup_default_branch(project)
    return unless project.mercurial?

    existing = project.default_branch
    return if existing.nil? || !existing.start_with?('topic/')

    project.repository.expire_root_ref_cache
    new_default = project.reload_default_branch
    return if new_default == existing

    protection = ProtectedBranch.find_by(project: project, name: existing)
    protection.destroy unless protection.nil?
  end
  # rubocop: enable CodeReuse/ActiveRecord

  def process_project_changes(post_received, project)
    user = identify_user(post_received)

    return false unless user

    push_options = post_received.push_options
    changes = post_received.changes

    # We only need to expire certain caches once per push
    expire_caches(post_received, project.repository)
    hg_fixup_default_branch(project)
    enqueue_project_cache_update(post_received, project)

    process_ref_changes(project, user, push_options: push_options, changes: changes)
    update_remote_mirrors(post_received, project)
    after_project_changes_hooks(project, user, changes.refs, changes.repository_data)
  end

  def process_wiki_changes(post_received, wiki)
    user = identify_user(post_received)
    return false unless user

    # We only need to expire certain caches once per push
    expire_caches(post_received, wiki.repository)
    wiki.repository.expire_statistics_caches

    ::Git::WikiPushService.new(wiki, user, changes: post_received.changes).execute
  end

  def process_snippet_changes(post_received, snippet)
    user = identify_user(post_received)

    return false unless user

    replicate_snippet_changes(snippet)

    expire_caches(post_received, snippet.repository)
    snippet.touch
    Snippets::UpdateStatisticsService.new(snippet).execute
  end

  def replicate_snippet_changes(snippet)
    # Used by Gitlab Geo
  end

  # Expire the repository status, branch, and tag cache once per push.
  def expire_caches(post_received, repository)
    repository.expire_status_cache if repository.empty?
    repository.expire_branches_cache if post_received.includes_branches?
    repository.expire_caches_for_tags if post_received.includes_tags?
  end

  # Schedule an update for the repository size and commit count if necessary.
  def enqueue_project_cache_update(post_received, project)
    stats_to_invalidate = [:repository_size]
    stats_to_invalidate << :commit_count if post_received.includes_default_branch?

    ProjectCacheWorker.perform_async(
      project.id,
      [],
      stats_to_invalidate,
      true
    )
  end

  def process_ref_changes(project, user, params = {})
    return unless params[:changes].any?

    Git::ProcessRefChangesService.new(project, user, params).execute
  end

  def update_remote_mirrors(post_received, project)
    return unless post_received.includes_branches? || post_received.includes_tags?

    return unless project.has_remote_mirror?

    project.mark_stuck_remote_mirrors_as_failed!
    project.update_remote_mirrors
  end

  def after_project_changes_hooks(project, user, refs, changes)
    repository_update_hook_data = Gitlab::DataBuilder::Repository.update(project, user, changes, refs)
    SystemHooksService.new.execute_hooks(repository_update_hook_data, :repository_update_hooks)
    Gitlab::UsageDataCounters::SourceCodeCounter.count(:pushes)
    emit_snowplow_event(project, user)
  end

  def log(message)
    Gitlab::GitLogger.error("POST-RECEIVE: #{message}")
  end

  def emit_snowplow_event(project, user)
    return unless Feature.enabled?(:route_hll_to_snowplow_phase2, project.namespace)

    Gitlab::Tracking.event(
      'PostReceive',
      'source_code_pushes',
      project: project,
      namespace: project.namespace,
      user: user
    )
  end
end

PostReceive.prepend_mod_with('PostReceive')
