# Mercurial projects in Heptapod

The primary objective of Heptapod is to add Mercurial support to GitLab.

## Version Control System (VCS) types

Each project has a `vcs_type` field to describe the VCS in use. Administrators
can select which of these are allowed in the application settings
(Admin Area > Settings > General > Visibility and access controls).


We are currently using two distinct VCS types for Mercurial projects, to support our
effort towards "native" handling of Mercurial.

### "Mercurial" VCS type

Value of the `vcs_type` field: `hg`.

This has been the default VCS type for new Projects since Heptapod 0.25 and was displayed
as "Mercurial (native)" before Heptapod 0.28.

Projects created right now under this mode should not need heavy and risky migrations
in the future. They are often referred to as "native Mercurial projects" in text resources,
such as documentation and issues.

The native Mercurial projects are exposed to the various GitLab components via the HGitaly
service. As of Heptapod 0.28, content is still converted to Git by default, but this can be
disabled by feature flags.

There are several stages of development and corresponding running modes:

+ HGitaly1 (past stage): the corresponding mode was completely removed in Heptapod 0.29.
+ HGitaly2 (current as of Heptapod 0.29): almost all Mercurial content is served via HGitaly,
  with very few exceptions, such as repository language statistics.

  The conversion to auxiliary Git repository still happens as a safety net, but can be disabled
  by feature flags, for advance evaluation.
+ HGitaly3 (most of the development still to be done): all read and write operations,
  including creation and maintenance tasks happen through HGitaly.
  The web application does not need to access the repositories directly,
  and the HGitaly servers can be remote.

More insight on these different ways to handle Mercurial is available
[on the Heptapod web site](https://heptapod.net/the-road-to-fully-native-mercurial-in-heptapod.md#the-road-to-fully-native-mercurial-in-heptapod)

### "Mercurial (legacy with Git conversion)" VCS type

Value of the `vcs_type` field: `hg_git`.

This VCS type was disabled by default for Project creation in Heptapod 0.25, but can
be enabled again in the application setttings.

This is the historical way to handle Mercurial repositories, exposing to the rest of
GitLab by maintaining a conversion to Git. Even though the server side Mercurial
repository is the single source of truth, this has many drawbacks.

The Rails (web) application accesses content for display through Gitaly, but also needs
to perform operations (such as merges) on the local file system. Hence the Gitaly server
*must* be local.

This does not prevent using also remote Gitaly servers, as long as they serve Git
projects only.

## Managing the native modes

In all cases, only native Mercurial projects are affected by this.

Warning: as of Heptapod 0.29, the mode without Git conversion is experimental,
see near the end how to reenable the Git conversion if needed.

#### The feature flags

- `hg_without_git`: prevents conversion to Git
  unless the Project explicitely requires it, e.g, to mirror-push to a Git
  server
- `hg_with_git_fallback`: make a given "fully native" project still convert
  content to Git, even if `hg_without_git` is enabled for the whole instance.

#### For the whole instance

To stop Git conversion for all native Mercurial projects do this under the [Rails console]:

```ruby
Feature.enable(:hg_without_git)
```

Disabling the mode without Git globally is **dangerous**, as you will need to catch up on the conversion
for all Projects that were pushed to since the feature flag was activated. Still, the command would be:

```ruby
Feature.disable(:hg_without_git)  # dangerous, see right above
```

It is not possible to use `Feature.disable()` for a single project.

To keep on converting for a given Project even if `hg_without_git` is
globally enabled (assuming no changeset has been pushed in the mode
without Git yet, otherwise see the dedicated section below):

```ruby
project = Project.find(project_id)  # project_id is the numeric ID,
Feature.enable?(:hg_with_git_fallback, project)  # only if no activity since global activation
```

To enable the mode without Git globally, except on a few projects, it is recommended
to set `hg_with_git_fallback` on them before the global activation of `hg_without_git`.

Of course, you are free to use other means to retrieve a `Project` instance, using
the `ActiveRecord` API (the Object Relational Mapping in use).

#### For a given project

##### Introspection

To know which mode a given project is running under, retrieve its
details using the [Projects API](../api/projects.md).

Here are the interesting fields about the kind of Mercurial support:

- `vcs_type`. Possible values:
   + `hg`: native Mercurial
   + `hg_git`: Mercurial (legacy with Git conversion)
   + `git` (for completeness)
- `hg_without_git`: if `true` no conversion to the Git auxiliary repo occurs

##### Using the Rest API

This is perhaps the simplest. Use the [Features flags API](../api/features.md#set-or-create-a-feature)
with this data:

- `name`: `hg_without_git`
- `value`: `true` or `false` whether you want to enable or disable
- `project`: full web path to the project

Notes:

- disabling the mode without Git does not work if it is enabled for the
  whole instance. Enable the `hg_with_git_fallback` feature flag instead.
- feature flag changes can take up to 60 seconds to be fully propagated. Automated agents relying on the
  API have to take this into account.

##### Under the [Rails console]:

First, take note of the project numeric ID. It is displayed on the project home page,
right under the project name.

To activate the mode without Git conversion on a single project:

```ruby
project = Project.find(project_id)  # project_id is the numeric ID,
Feature.enable(:hg_without_git, project)
```

To disable the mode without Git conversion on a single project, if the feature flag is not globally set
to `true`:

```ruby
project = Project.find(project_id)  # project_id is the numeric ID,
Feature.disable(:hg_without_git, project)
```

To disable the mode without Git conversion on a single project, if the feature flag is globally
set to `true`:

```ruby
project = Project.find(project_id)  # project_id is the numeric ID,
Feature.enable(:hg_with_git_fallback, project)
```

#### How to resume Git conversion if it has been stopped by feature flags

If you want to *resume* conversion to Git and it has been
stopped for some time, you need to

- (recommended) stop accepting changes by archiving the relevant
  Projects
- run `hg hpd-git-resync` for the given projects, following
  [how to run a hg command on the server](https://heptapod.net/pages/faq.html#hg-command)
- flip the feature flags (either the global one or the per-project) as
  explained above
- unarchive the project if you did archive it.


[Rails console]: operations/rails_console.md




