# frozen_string_literal: true

module Backup
  # Backup and restores repositories using gitaly-backup
  #
  # gitaly-backup can work in parallel and accepts a list of repositories
  # through input pipe using a specific json format for both backup and restore
  class GitalyBackup
    # @param [StringIO] progress IO interface to output progress
    # @param [Integer] max_parallelism max parallelism when running backups
    # @param [Integer] storage_parallelism max parallelism per storage (is affected by max_parallelism)
    def initialize(progress, max_parallelism: nil, storage_parallelism: nil, incremental: false, backup_id: nil)
      @progress = progress
      @max_parallelism = max_parallelism
      @storage_parallelism = storage_parallelism
      @incremental = incremental
    end

    def start(type, backup_repos_path, backup_id: nil)
      raise Error, 'already started' if started?

      if type == :create && !incremental?
        FileUtils.rm_rf(backup_repos_path)
      end

      command = case type
                when :create
                  'create'
                when :restore
                  'restore'
                else
                  raise Error, "unknown backup type: #{type}"
                end

      args = ['-layout', 'pointer']
      args += ['-parallel', @max_parallelism.to_s] if @max_parallelism
      args += ['-parallel-storage', @storage_parallelism.to_s] if @storage_parallelism
      if type == :create
        args += ['-incremental'] if incremental?
        args += ['-id', backup_id] if backup_id
      end

      @input_stream, stdout, @thread = Open3.popen2(build_env, bin_path, command, '-path', backup_repos_path, *args)

      @out_reader = Thread.new do
        IO.copy_stream(stdout, @progress)
      end
    end

    def finish!
      return unless started?

      @input_stream.close
      [@thread, @out_reader].each(&:join)
      status =  @thread.value

      @thread = nil

      raise Error, "gitaly-backup exit status #{status.exitstatus}" if status.exitstatus != 0
    end

    def enqueue(container, repo_type)
      raise Error, 'not started' unless started?

      repository =
        if repo_type.project? # TODO find better and more universal solution
          container.repository
        elsif repo_type.wiki?
          container.wiki.repository
        else
          repo_type.repository_for(container)
        end

      schedule_backup_job(repository, always_create: repo_type.project?)
    end

    private

    # TODO make a standard method of Repository or check how previous backup code is coping
    def repo_with_git?(repo)
      # if the question cannot even be asked, it means this is
      # a Git repository.  Using respond_to? to survive foreseeable
      # refactorings
      return true unless repo.respond_to?(:hg_with_git?)

      repo.hg_with_git?
    end

    # TODO same remark as in the Git case
    def repo_with_mercurial?(repo)
      repo.respond_to?(:hg_relative_path)
    end

    def incremental?
      @incremental
    end

    # Schedule a new backup job through a non-blocking JSON based pipe protocol
    #
    # @see https://gitlab.com/gitlab-org/gitaly/-/blob/master/doc/gitaly-backup.md
    # early return ("guard clause") not fit for succession of conditions
    # rubocop:disable Style/GuardClause
    def schedule_backup_job(repository, **named_args)
      raw = repository.raw
      if repo_with_git?(raw)
        schedule_backup_job_git(repository, **named_args)
      end

      if repo_with_mercurial?(raw)
        schedule_backup_job_mercurial(raw, **named_args)
      end
    end
    # rubocop:enable Style/GuardClause

    def schedule_backup_job_git(repository, always_create:)
      storage = repository.storage
      connection_params = Gitlab::GitalyClient.connection_data(storage)

      json_job = {
        address: connection_params['address'],
        token: connection_params['token'],
        storage_name: repository.storage,
        relative_path: repository.relative_path,
        gl_project_path: repository.gl_project_path,
        always_create: always_create
      }.to_json

      @input_stream.puts(json_job)
    end

    def schedule_backup_job_mercurial(repository, always_create:)
      storage = Gitlab::GitalyClient.vcs_qualified_storage(repository.storage, 'hg')
      connection_params = Gitlab::GitalyClient.connection_data(storage)

      json_job = {
        address: connection_params['address'],
        token: connection_params['token'],
        storage_name: repository.storage,
        relative_path: repository.hg_relative_path,
        gl_project_path: repository.gl_project_path,
        always_create: always_create
      }.to_json

      @input_stream.puts(json_job)
    end

    def build_env
      {
        'SSL_CERT_FILE' => Gitlab::X509::Certificate.default_cert_file,
        'SSL_CERT_DIR' => Gitlab::X509::Certificate.default_cert_dir
      }.merge(ENV)
    end

    def started?
      @thread.present?
    end

    def bin_path
      raise Error, 'gitaly-backup binary not found and gitaly_backup_path is not configured' unless Gitlab.config.backup.gitaly_backup_path.present?

      File.absolute_path(Gitlab.config.backup.gitaly_backup_path)
    end
  end
end
