# frozen_string_literal: true

module Gitlab
  module GitalyClient
    class MercurialRepositoryService
      include Gitlab::EncodingHelper

      # gRPC enums are converted to symbols upon read-back
      # See https://protobuf.dev/reference/ruby/ruby-generated/#enum
      AUTO_PUBLISH_MAPPING = { WITHOUT_TOPIC: "without-topic",
                               ALL: "all",
                               NOTHING: "nothing" }.freeze
      AUTO_PUBLISH_REVERSE_MAPPING = AUTO_PUBLISH_MAPPING.to_a.to_h(&:reverse).freeze

      # 'repository' is a Mercurial subclass of Gitlab::Git::Repository
      def initialize(repository)
        @repository = repository
        @gitaly_repo = repository.gitaly_repository
        @storage = Gitlab::GitalyClient.vcs_qualified_storage(
          repository.storage, 'hg')
      end

      def config_item_bool(section, item)
        request = Hgitaly::GetConfigItemRequest.new(
          repository: @gitaly_repo,
          section: section,
          name: item,
          as_type: Hgitaly::ConfigItemType::BOOL
        )
        response = GitalyClient.call(
          @storage, :mercurial_repository_service, :get_config_item, request,
          timeout: GitalyClient.fast_timeout
        )
        response.as_bool
      end

      def get_managed_config(local: false)
        request = Hgitaly::GetManagedConfigRequest.new(
          repository: @gitaly_repo,
          local: local
        )
        response = GitalyClient.call(
          @storage, :mercurial_repository_service, :get_managed_config, request,
          timeout: GitalyClient.fast_timeout
        )

        section = {}
        hepta = response.heptapod

        section['allow-bookmarks'] = hepta.allow_bookmarks if hepta.has_allow_bookmarks?
        section['allow-multiple-heads'] = hepta.allow_multiple_heads if hepta.has_allow_multiple_heads?
        section['auto-publish'] = AUTO_PUBLISH_MAPPING[hepta.auto_publish] if hepta.has_auto_publish?
        [response.inherit, section]
      end

      def set_managed_config(inherit, heptapod, remove_items, by_line)
        fields = {}
        book = heptapod['allow-bookmarks']
        fields[:allow_bookmarks] = book unless book.nil?
        multi_heads = heptapod['allow-multiple-heads']
        fields[:allow_multiple_heads] = multi_heads unless multi_heads.nil?
        auto_pub = heptapod['auto-publish']
        fields[:auto_publish] = AUTO_PUBLISH_REVERSE_MAPPING[auto_pub] unless auto_pub.nil?

        request = Hgitaly::SetManagedConfigRequest.new(
          repository: @gitaly_repo,
          by_line: by_line,
          heptapod: Hgitaly::HeptapodConfigSection.new(**fields)
        )
        request.inherit = inherit unless inherit.nil?

        GitalyClient.call(
          @storage, :mercurial_repository_service, :set_managed_config, request,
          timeout: GitalyClient.fast_timeout
        )
      end

      def push(remote_url, only_branches_matching: [], ssh_key: nil, known_hosts: nil, include_drafts: false)
        remote_peer = Hgitaly::MercurialPeer.new(
          url: remote_url,
          ssh_key: ssh_key,
          ssh_known_hosts: known_hosts
        )
        request = Hgitaly::PushRequest.new(
          repository: @gitaly_repo,
          remote_peer: remote_peer,
          only_gitlab_branches_matching: only_branches_matching,
          include_drafts: include_drafts
        )
        response = GitalyClient.call(
          @storage, :mercurial_repository_service, :push, request,
          timeout: GitalyClient.long_timeout
        )
        response.new_changesets
      end
    end
  end
end
