# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# frozen_string_literal: true

module Gitlab
  module Utils
    module LineEndings
      InconsistentLineEndings = Class.new(StandardError)

      UNIX = "\n"
      WINDOWS = "\r\n"
      HTML_TEXTAREA = WINDOWS
      MAC_CLASSIC = "\r" # MacOS before OSX

      extend self

      # Analyze content for consistent line endings
      #
      # returns
      # - the line ending (`UNIX`, `WINDOWS`, `MAC_CLASSIC`) if it is consistently
      #   used in `content`,
      # - :mixed if several distinct line endings are in use
      # - or nil if there is no line ending
      #
      # If last line doesn't have an End-of-Line code, it doesn't count as an inconsistency
      def analyze(content)
        res = nil
        lines = content.lines
        return if lines.empty?

        # ordering matters, because we'll use early breaks in early loops
        # keeping as litterals instead of constants to make what happens clear
        eols = ["\r\n", "\n", "\r"]
        content.lines.each do |line|
          eols.each do |eol|
            # special cases for MAC_CLASSIC due to Ruby line spliting not recognizing it.
            # this eol is checked last. With the `break` for other eols, include? is enough.
            if eol == MAC_CLASSIC && line.include?(eol)
              res ||= eol
              return :mixed if res != eol
            elsif line.ends_with?(eol)
              # chomp considers \r to be a line ending, but it only removes the last
              # line ending, so could be usable. Still, that makes a whole copy
              # that we'd better avoid
              return :mixed if line[..-eol.length - 1].include?(MAC_CLASSIC)

              if res.nil?
                res = eol
              elsif res != eol
                return :mixed
              end

              break
            end
          end
        end
        res
      end

      # Return a version of `content` with line endings as in as in `reference_content`
      #
      # Limitations: does not understand MAC_CLASSIC line endings in `content` (not needed
      #              for current applications as of this writing)
      #
      # Param `default`: can be used to specify a default line ending in case `reference_content`
      #     has no line endings.
      #
      # Param `solve_inconsistency_as`: if `reference_content` has inconsistent line endings,
      #     this is an end of line to choose instead. If not given, `InconsistentLineEndings`
      #     is raised in that situation.
      def replace_as_in(content, reference_content, default: UNIX, solve_inconsistency_as: nil)
        ref_eol = analyze(reference_content)
        if ref_eol == :mixed
          raise InconsistentLineEndings if solve_inconsistency_as.nil?

          ref_eol = solve_inconsistency_as
        end

        # could not find how to join iterators without collecting first in an Array
        # special case: if last line endswith '\r' the chomp will remove it because
        # line splitting and chomp aren't consistent. Is it really worth
        content.each_line.map(&:chomp).to_a.join(ref_eol || default)
      end
    end
  end
end
