# frozen_string_literal: true

module Heptapod
  module NativeMercurial
    # Native Mercurial Migrator
    #
    # This is responsible for migrating a mercurial project from hg_git backend to native hg.
    class Migrator
      SHA_MAX_LEN = 40
      SHA_REGEX = /\b[0-9a-f]{7,40}\b/.freeze
      EXCLUSIVE_LEASE_KEY_SEGMENT = 'project_migrate_hg_native'
      EXCLUSIVE_LEASE_TIMEOUT = 2.minutes.to_i
      EXCLUSIVE_LEASE_NOTES_RENEWAL_RATE = 10 # renew the lease every N notes

      # user: the user triggering the migration, for subsystems that require it
      def initialize(project, user)
        @project = project
        @user = user
      end

      # Tell whether migration to native is currently running for given project
      # For the time being, we rely on ExclusiveLeave, hence on Redis. We might
      # later on introduce a more persistent marker (SQL database)
      def self.project_migrating?(project)
        new(project, nil).exclusive_lease_taken?
      end

      def exclusive_lease_key
        "#{EXCLUSIVE_LEASE_KEY_SEGMENT}:#{@project.id}"
      end

      # Tell whether the exclusive lease is taken, possibly by another instance
      def exclusive_lease_taken?
        # Passing a non-nil uuid value just to avoid useless SecureRandom.uuid computation,
        # as we don't want to take or renew a lease
        Gitlab::ExclusiveLease.new(exclusive_lease_key, uuid: "", timeout: 1).exists?
      end

      def timeless_save(record)
        Gitlab::Timeless.timeless(record, &:save!)
      end

      def vcs_type_enforcer
        proj_initial_vcs = @project.vcs_type
        if @project.vcs_type == 'hg'
          @project.vcs_type = 'hg_git'
          timeless_save(@project)
        end

        yield
      ensure
        # ensure that native project remains native in case of errors
        if proj_initial_vcs == 'hg'
          @project.vcs_type = 'hg'
          timeless_save(@project)
        end
      end

      def with_project_readonly
        ::Projects::UpdateService.new(@project, @user, archived: true).execute
        yield

        ::Projects::UpdateService.new(@project, @user, archived: false).execute
      end

      def exclusive_lease
        @exclusive_lease ||= Gitlab::ExclusiveLease.new(exclusive_lease_key, timeout: EXCLUSIVE_LEASE_TIMEOUT)
      end

      def release_exclusive_lease(uuid)
        Gitlab::ExclusiveLease.cancel(exclusive_lease_key, uuid)
      end

      def renew_exclusive_lease
        @exclusive_lease.renew unless @exclusive_lease.nil?
      end

      def try_obtain_exclusive_lease
        uuid = exclusive_lease.try_obtain
        unless uuid
          log_warning("Cannot obtain an exclusive lease for native Mercurial migration "\
                      "of Project id=#{@project.id}. Migration must be already running.")
          return
        end

        begin
          yield
        ensure
          release_exclusive_lease(uuid)
        end
      end

      def init_sha_map
        @hg_sha_map = @project.repository.hg_sha_map
      end

      def running?
        Heptapod::NativeMigrationLog.running?(@project)
      end

      def with_record
        raise StandardError, "A migration for Project id=#{project.id} is currently running" if running?

        record = Heptapod::NativeMigrationLog.start(@project)
        record.save!

        yield record

      ensure
        # We must keep track of all failed attempts, finished does not mean
        # succesful.
        record.state = :finished
        record.save!
      end

      # take care of all startup initializations and final cleanups,
      # like a Python context manager would.
      def enter
        res = nil
        try_obtain_exclusive_lease do
          with_project_readonly do
            vcs_type_enforcer do
              with_record do |rec|
                res = yield rec
              end
            end
          end
        end
        res
      end

      def migrate!
        enter do |record|
          init_sha_map
          do_migrate(record)
        end
      end

      def log_error(err, msg)
        Gitlab::AppLogger.error(
          message: msg,
          backtrace: Gitlab::BacktraceCleaner.clean_backtrace(err.backtrace),
          project_id: @project.id,
          project_full_path: @project.full_path
        )
      end

      def log_warning(msg)
        Gitlab::AppLogger.warn(
          message: msg,
          project_id: @project.id,
          project_full_path: @project.full_path
        )
      end

      # ensure all gitlab specific state files are present
      def hg_ensure_state_files
        renew_exclusive_lease
        output, status = @project.repository.hg_call(
          ["hpd-ensure-all-gitlab-specific-state-files"], force_system_user: true
        )
        raise StandardError, output.strip unless status == 0
      end

      def set_final_vcs_type
        @project.vcs_type = 'hg'
        timeless_save(@project)
      end

      # rubocop: disable CodeReuse/ActiveRecord
      # Automatically set the hg_legacy_git_conversion setting unless it has already been set
      #
      # We allow users to set it even on `hg_git` projects, so that they can do it in advance to avoid
      # any disruption when going native.
      def set_hg_legacy_git_conversion
        @project.hg_legacy_git_conversion ||= !RemoteMirror.find_by(project_id: @project.id, hg_mirror_type: nil).nil?
      end
      # rubocop: enable CodeReuse/ActiveRecord

      # perform the migration, returning a mapping of errors per type
      #
      # Most errors will raise an exception instead of being added to
      # the return.
      def do_migrate(record)
        Gitlab::AppLogger.info(
          message: "Starting native hg migration",
          project_id: @project.id,
          project_full_path: @project.full_path
        )
        errors = {}

        hg_ensure_state_files
        migrate_merge_requests
        record.rollbackable = false

        set_hg_legacy_git_conversion
        migrate_issues
        migrate_pipelines
        migrate_releases
        migrate_notes(errors)
        record.rollbackable = false
        set_final_vcs_type

        record.migration_errors = errors
        errors
      rescue StandardError => err
        msg = "Native mercurial migration failed: #{err}"
        log_error(err, msg)
        raise
      ensure
        set_final_vcs_type unless record.rollbackable
        invalidate_renderings
      end

      # rubocop: disable CodeReuse/ActiveRecord
      def invalidate_renderings
        Note.where(project: @project).update_all(note_html: nil)
        MergeRequest.where(project: @project).update_all(description_html: nil,
                                                         title_html: nil)
        Issue.where(project: @project).update_all(description_html: nil,
                                                  title_html: nil)
        Release.where(project: @project).update_all(description_html: nil)
      end
      # rubocop: enable CodeReuse/ActiveRecord

      def migrate_issues
        @project.issues.each do |issue|
          renew_exclusive_lease
          issue.title = replace_git_shas_with_hg(issue.title)
          issue.description = replace_git_shas_with_hg(issue.description)
          timeless_save(issue)
        end
      end

      def migrate_pipelines
        @project.all_pipelines.each do |pline|
          renew_exclusive_lease
          pline.sha = replace_git_shas_with_hg(pline.sha)
          pline.before_sha = replace_git_shas_with_hg(pline.before_sha)
          pline.source_sha = replace_git_shas_with_hg(pline.source_sha)
          pline.target_sha = replace_git_shas_with_hg(pline.target_sha)
          timeless_save(pline)
        end
      end

      # Note fields whose change should trigger a database save
      def note_watched_fields(note)
        [
          note.note, note.commit_id,
          note.original_position, note.position, note.change_position
        ]
      end

      def save_note_if_changed(note)
        before = note_watched_fields(note)
        yield
        after = note_watched_fields(note)

        return if after == before

        note.note_html = nil
        note.importing = true
        note.hg_native_migrating = true
        timeless_save(note)
      end

      def migrate_notes(errors)
        count = 0
        failed = []
        @project.notes.each do |note|
          count += 1
          renew_exclusive_lease if count % EXCLUSIVE_LEASE_NOTES_RENEWAL_RATE == 0
          save_note_if_changed(note) do
            note.note = replace_git_shas_with_hg(note.note)
            migrate_note_positions(note)
            note.commit_id = replace_git_shas_with_hg(note.commit_id)
          end
        rescue StandardError => err
          log_error(err, "Failed to migrate Note id=#{note.id}")
          failed << note.id
        end
        errors[:notes] = failed unless failed.empty?
      end

      def migrate_note_positions(note)
        _migrate_note_position_var(note, :original_position)
        _migrate_note_position_var(note, :position)
        _migrate_note_position_var(note, :change_position)
      end

      def _migrated_position(pos)
        pos_attrs = pos.to_h
        pos_attrs[:base_sha] = replace_git_shas_with_hg(pos_attrs[:base_sha])
        pos_attrs[:start_sha] = replace_git_shas_with_hg(
          pos_attrs[:start_sha])
        pos_attrs[:head_sha] = replace_git_shas_with_hg(pos_attrs[:head_sha])
        Gitlab::Diff::Position.new(pos_attrs)
      end

      # rubocop:disable GitlabSecurity/PublicSend
      # migrate the position instance variable `name` on `note`
      # to call with hardcoded constants only
      def _migrate_note_position_var(note, name)
        value = note.public_send(name)
        return unless value.instance_of? Gitlab::Diff::Position

        note.public_send("#{name}=", _migrated_position(value))
      end
      # rubocop:enable GitlabSecurity/PublicSend

      def migrate_releases
        @project.releases.each do |rl|
          renew_exclusive_lease
          rl.sha = replace_git_shas_with_hg(rl.sha)
          rl.description = replace_git_shas_with_hg(rl.description)
          timeless_save(rl)
        end
      end

      def migrate_merge_requests
        @project.merge_requests.each do |mr|
          renew_exclusive_lease
          mr.title = replace_git_shas_with_hg(mr.title)
          mr.title_html = nil
          mr.description = replace_git_shas_with_hg(mr.description)

          # update all the columns ending with '_sha'
          mr.merge_commit_sha = replace_git_shas_with_hg(mr.merge_commit_sha)
          mr.rebase_commit_sha = replace_git_shas_with_hg(mr.rebase_commit_sha)
          mr.in_progress_merge_commit_sha = (
            replace_git_shas_with_hg(mr.in_progress_merge_commit_sha)
          )
          mr.squash_commit_sha = replace_git_shas_with_hg(mr.squash_commit_sha)
          mr.merge_ref_sha = replace_git_shas_with_hg(mr.merge_ref_sha)
          migrate_merge_request_diffs(mr)
          timeless_save(mr)
        end
      end

      # rubocop: disable CodeReuse/ActiveRecord
      def migrate_merge_request_diffs(mr)
        mr.merge_request_diffs.each do |diff|
          diff.base_commit_sha = replace_git_shas_with_hg(diff.base_commit_sha)
          diff.head_commit_sha = replace_git_shas_with_hg(diff.head_commit_sha)
          diff.start_commit_sha = replace_git_shas_with_hg(diff.start_commit_sha)
          timeless_save(diff)
        end

        # The table for MergeRequestDiffCommit has a composite primary key:
        # `PRIMARY KEY, btree (merge_request_diff_id, relative_order)`,
        # preventing updates using the normal ActiveRecord API, that only
        # supports single column primary keys.
        # Hence we have to go low level and build a query by hand with Arel,
        # the SQL query builder and AST representation used by ActiveRecord.
        # The crucial fact here is that the WHERE clauses pin down the two
        # columns that are part of the primary key (as of this writing).
        arel_table = MergeRequestDiffCommit.arel_table
        mr.all_commits.each do |mr_diff_commit|
          hg_sha = replace_git_shas_with_hg(mr_diff_commit.sha)
          hg_sha = "\\x#{hg_sha}"
          query = Arel::UpdateManager.new
            .table(arel_table)
            .set([[arel_table[:sha], hg_sha]])
            .where(arel_table[:merge_request_diff_id].eq(mr_diff_commit.merge_request_diff_id))
            .where(arel_table[:relative_order].eq(mr_diff_commit.relative_order))
            .to_sql
          MergeRequestDiffCommit.connection.execute(query)
        end
      end
      # rubocop: enable CodeReuse/ActiveRecord

      private

      def replace_git_shas_with_hg(text)
        return unless text.present? && @hg_sha_map.present?

        git_shas = @hg_sha_map.keys
        text.gsub(SHA_REGEX) do |match|
          if match.length == SHA_MAX_LEN && git_shas.include?(match)
            @hg_sha_map[match]
          else
            full_git_sha = lambda {
              git_shas.each do |sha|
                return sha if sha.starts_with? match
              end
              nil
            }.call
            full_git_sha.nil? ? match : @hg_sha_map[full_git_sha]
          end
        end
      end
    end
  end
end
