# frozen_string_literal: true

namespace :heptapod do
  namespace :experimental do
    desc 'GitLab | Migrate to Native Hg'
    task :hg_migrate_native, [:project_id, :force, :username] => [:gitlab_environment] do |task, args|
      args.with_defaults(force: false, username: 'root')
      project = Project.find(args[:project_id])
      user = User.find_by(username: args[:username])

      if user.nil?
        puts("abort: unknown user #{args[:username].inspect}. "\
             "Please pass an existing username with enough rights explicitely".color(:red))
        exit(1)
      end

      unless project.mercurial?
        puts "abort: project is not a mercurial project".color(:red)
        exit(1)
      end

      unless user.can?(:admin_project, project)
        puts("abort: user #{user.username} does not have "\
             "administration rights on project id=#{project.id}".color(:red))
        exit(1)
      end

      if project.vcs_type == 'hg' && !args[:force]
        puts "abort: project is already migrated, use 'force' to migrate anyway".color(:red)
        exit(1)
      end

      check_gitaly_connection
      check_hgitaly_connection
      migrator = Heptapod::NativeMercurial::Migrator.new(project, user)
      errors = migrator.migrate!
      if errors.each_value.all?(&:empty?)
        puts 'done!'.color(:green)
      else
        puts "there are errors for #{errors}. Please check application log for details".color(:yellow)
      end
    end
  end
end

def check_gitaly_connection
  storage_name = nil
  Gitlab.config.repositories.storages.each do |name, _details|
    storage_name = name
    Gitlab::GitalyClient::ServerService.new(name).info
  end
rescue GRPC::Unavailable => ex
  puts "Failed to connect to Gitaly...".color(:red)
  puts "Storage name: #{storage_name}"
  puts "Error: #{ex}"
  exit 1
end

def check_hgitaly_connection
  storage_name = nil
  Gitlab.config.repositories.storages.each do |name, _details|
    storage_name = name
    storage = Gitlab::GitalyClient.vcs_qualified_storage(name, 'hg')
    Gitlab::GitalyClient::ServerService.new(storage).info
  end
rescue GRPC::Unavailable => ex
  puts "Failed to connect to HGitaly...".color(:red)
  puts "Storage name: #{storage_name}"
  puts "Error: #{ex}"
  exit 1
end
