#!/bin/sh

set -e

# determination of corresponding Omnibus branch, even for intermediate
# releases (gives something to check in the log if not actually in use)

# why three `if` instead of a defaulting `else` case? Beacuse we want a
# closed list of admissible series, to avoid painful
# release commit and tag creation in the wrong branch of omninus-heptapod
# if, e.g.,  we change the name of the CI image and forget to update this
# script.
if [ "${HEPTAPOD_SERIES}" = "oldstable" ]; then
    TRIGGER_OMNIBUS_BRANCH=heptapod-oldstable
elif [ "${HEPTAPOD_SERIES}" = "stable" ]; then
    TRIGGER_OMNIBUS_BRANCH=heptapod-stable
elif [ "${HEPTAPOD_SERIES}" = "default" ]; then
    TRIGGER_OMNIBUS_BRANCH=heptapod
else
    echo "Could not determine Omnibus branch for automatic release." >&2
    echo "Value of HEPTAPOD_SERIES: '${HEPTAPOD_SERIES}'" >&2
    echo "This must be fixed immediately, even if the present pipeline is" >&2
    echo "not for a tag (in which case next release is in great danger)." >&2
    echo "Aborting." >&2
    exit 1
fi

echo "Target Omnibus branch for automatic release creation: $TRIGGER_OMNIBUS_BRANCH"

if echo "$CI_COMMIT_TAG" | grep '^heptapod-' > /dev/null; then
    TARBALL=$(echo -n $CI_COMMIT_TAG | sed s/heptapod/heptapod-rails/g)
    # pulling on the current branch and the two main heptapod ones
    # so that `hg archive` should see the tag (very unlikely the tagging
    # changeset would be elsewhere, not worth pulling the whole repo)
    # using `hg log` instead of `hg branch` to avoid newline for safety
    hg pull -q -r heptapod -r heptapod-stable -r heptapod-oldstable -r `hg log -r. -T "{branch}"` ${CI_PROJECT_URL} || /bin/true
    SUBDIR=public
else
    TARBALL=heptapod-rails-$(hg log -r. -T "{node}")
    SUBDIR=public/intermediate
    # disable automatic Omnibus commit creation
    TRIGGER_OMNIBUS_BRANCH=
fi

# it's nicer for console output to add the extension right away
TARBALL=${TARBALL}.tar.gz

TARBALL_DIR=/tmp
TARBALL_PATH=${TARBALL_DIR}/${TARBALL}

# Directory used as artifact to pass versions to downstream Omnibus jobs,
# without committing them in the Omnibus repository.
# Make sure it exists and no previous build remnants will interfere
VERSIONS_MANIFEST=tmp/heptapod_tarball_job_versions
rm -rf ${VERSIONS_MANIFEST}
mkdir ${VERSIONS_MANIFEST}

echo "Generating ${TARBALL}"
hg archive -t tgz ${TARBALL_PATH}
echo "Generating SHA-256 checksum"
cd ${TARBALL_DIR}
sha256sum ${TARBALL} > ${TARBALL_PATH}.sha256sum
cd -
cat ${TARBALL_PATH}.sha256sum


echo "Filling-in ${VERSIONS_MANIFEST}"
if [ -n "$CI_COMMIT_TAG" ]; then
    echo -n $CI_COMMIT_TAG | sed s/heptapod-//g > ${VERSIONS_MANIFEST}/HEPTAPOD_VERSION
else
    HEPTAPOD_MINOR_VERSION=`sed 's/^\([0-9]\+\)[.]\([0-9]\+\)[.].*$/\1.\2/g' HEPTAPOD_VERSION`
    echo ${HEPTAPOD_MINOR_VERSION}-${CI_COMMIT_HG_SHA} > ${VERSIONS_MANIFEST}/HEPTAPOD_VERSION
fi
cut -f1 -d ' ' ${TARBALL_PATH}.sha256sum > ${VERSIONS_MANIFEST}/HEPTAPOD_RAILS_SHA256
mkdir -p tmp/tests  # needed by rhgitaly-download
scripts/rhgitaly-download --sha256 > ${VERSIONS_MANIFEST}/RHGITALY_SHA256
cp HEPTAPOD_SHELL_VERSION ${VERSIONS_MANIFEST}/
cp RHGITALY_SERVER_VERSION ${VERSIONS_MANIFEST}/


# wrapping VERSIONS_MANIFEST in yet another tarball (see heptapod#699)
VERSIONS_TARBALL_PATH=/tmp/heptapod-rails-versions-${CI_COMMIT_HG_SHA}.tar
tar cf ${VERSIONS_TARBALL_PATH} ${VERSIONS_MANIFEST}

if [ -n "$HEPTAPOD_PACKAGING_GPG_PRIVKEY" ]; then
    `dirname $0`/heptapod-sign-package $HEPTAPOD_PACKAGING_GPG_PRIVKEY \
                $TARBALL_PATH
else
    echo "Key not present, skipping GPG signature. "
    echo "    This is normal on non-protected tags and branches"
fi

if [ -n "$HEPTAPOD_TARBALL_UPLOAD_KEY" ]; then
    `dirname $0`/heptapod-sftp-push \
                upload.heptapod.net \
                "$HEPTAPOD_TARBALL_UPLOAD_KEY" \
                "$TARBALL_PATH" \
                "${SUBDIR}/src"
    `dirname $0`/heptapod-sftp-push \
                upload.heptapod.net \
                "$HEPTAPOD_TARBALL_UPLOAD_KEY" \
                "$VERSIONS_TARBALL_PATH" \
                "${SUBDIR}/src"
fi

echo

HGRCPATH=/tmp/job.hgrc
export HGRCPATH
cat > $HGRCPATH <<EOF
[extensions]
topic=
evolve=

[ui]
username = "Heptapod Release Bot <heptapod-release.bot@heptapod.net>"
EOF

# we prefer the waste of cloning to tmp rather than an untested update script
HEPTAPOD_OMNIBUS_CLONE=/tmp/omnibus-heptapod
HEPTAPOD_OMNIBUS_REPO=foss.heptapod.net/heptapod/omnibus-heptapod

if [ -z "$HEPTAPOD_OMNIBUS_ACCESS_CREDS" ]; then
echo "No credentials to push to omnibus-heptapod repo."
fi

if [ -n "$TRIGGER_OMNIBUS_BRANCH" -a -n "$HEPTAPOD_OMNIBUS_ACCESS_CREDS" ]; then
    # here it would be nice to give remotefilelog a try
    # in this case we want to be 100% sure not to miss a tag, hence no -r
    hg clone -u ${TRIGGER_OMNIBUS_BRANCH} \
       https://${HEPTAPOD_OMNIBUS_REPO} \
       $HEPTAPOD_OMNIBUS_CLONE

    cp ${VERSIONS_MANIFEST}/HEPTAPOD_VERSION ${HEPTAPOD_OMNIBUS_CLONE}/
    cp ${VERSIONS_MANIFEST}/HEPTAPOD_RAILS_SHA256 ${HEPTAPOD_OMNIBUS_CLONE}/HEPTAPOD_RAILS_SHA256
    cp ${VERSIONS_MANIFEST}/HEPTAPOD_SHELL_VERSION ${HEPTAPOD_OMNIBUS_CLONE}/
    cp ${VERSIONS_MANIFEST}/RHGITALY_SERVER_VERSION ${HEPTAPOD_OMNIBUS_CLONE}/
    cp ${VERSIONS_MANIFEST}/RHGITALY_SHA256 ${HEPTAPOD_OMNIBUS_CLONE}/

    echo "Committing version files"
    hg commit -R ${HEPTAPOD_OMNIBUS_CLONE} -qm "Updated version files for release"
    echo "Done as:"
    hg export -R ${HEPTAPOD_OMNIBUS_CLONE} .

    # Omnibus maintains an additional -1 build number.
    OMNIBUS_TAG=${CI_COMMIT_TAG}-1
    echo "Attempting to create tag ${OMNIBUS_TAG}. In case of failure, "
    echo "especially because the tag already exists, please finish manually."
    hg tag -R ${HEPTAPOD_OMNIBUS_CLONE} ${OMNIBUS_TAG}
    echo "Done as:"
    hg log -R ${HEPTAPOD_OMNIBUS_CLONE} -r .

    hg push -R ${HEPTAPOD_OMNIBUS_CLONE} -r . \
       https://${HEPTAPOD_OMNIBUS_ACCESS_CREDS}@${HEPTAPOD_OMNIBUS_REPO}
fi
