# frozen_string_literal: true

require "spec_helper"

RSpec.describe Gitlab::Mercurial::HgGitRepository, :seed_helper do
  include Gitlab::EncodingHelper
  include RepoHelpers

  let(:mutable_repository) do
    repo = described_class.new('default', 'hg-git-mutable.git', '', 'group/project')
    repo.unit_tests_skip_hooks = true
    repo
  end

  let(:mutable_repository_path) do
    "#{File.join(TestEnv.repos_path, mutable_repository.relative_path).chomp('.git')}.hg"
  end

  let(:user) { build(:user) }
  let(:admin_user) { create(:admin) }
  let(:project) { create(:hg_git_project, :hg_repository) }
  let(:empty_repository) do
    repo = mutable_repository

    status = repo.create_repository(namespace_fs_path: "group",
                                    in_subgroup: false)
    expect(status).to be(true)
    repo
  end

  let(:non_empty_repository) do
    repo = empty_repository

    repo.commit_files(user,
                      branch_name: 'branch/default',
                      message: "Commit 0",
                      author_email: "jdoe1@heptapod.test",
                      author_name: "John Doe1",
                      actions: [{ action: :create,
                                  file_path: 'foo',
                                  content: "foo0\n" }])
    repo.expire_has_local_branches_cache
    expect(repo.empty?).to be(false)
    repo
  end

  let(:non_empty_repository_path) { mutable_repository_path }

  # viewing a repo through HGitaly will be useful for assertions
  # without resorting to the file system (easier to adapt in the future)
  def as_hgitaly_repo(hg_git_repo)
    Gitlab::Mercurial::HgitalyRepository.new(hg_git_repo.storage,
                                             hg_git_repo.relative_path,
                                             nil, nil)
  end

  def git_repo_path(hg_git_repo)
    File.join(TestEnv.repos_path, hg_git_repo.relative_path)
  end

  describe "Respond to" do
    subject { mutable_repository }

    # from the base class
    it { is_expected.to respond_to(:root_ref) }
    it { is_expected.to respond_to(:tags) }

    # specific
    it { is_expected.to respond_to(:hg_sha_map) }
  end

  describe "#hg_call" do
    subject { empty_repository }

    after do
      subject.remove
    end

    it "spawns hg process having heptapod.native=no with the default environment variables" do
      output, _status = subject.hg_call(['config', 'heptapod.native'])
      expect(output.strip).to eq("no")
    end

    it "spawns hg process having heptapod.native=no with the write environment variables" do
      output, _status = subject.hg_call(['config', 'heptapod.native'],
                                        env: subject.hg_env_for_write(user))
      expect(output.strip).to eq("no")
    end

    it "spawns hg process having heptapod.native=no with the minimal environment variables" do
      output, _status = subject.hg_call(['config', 'heptapod.native'],
                                        env: subject.hg_env_minimal)
      expect(output.strip).to eq("no")
    end
  end

  describe "overridden methods" do
    describe '#bundle_to_disk' do
      let(:repository) { non_empty_repository }
      let(:save_path) { File.join(Dir.tmpdir, "repo-#{SecureRandom.hex}.bundle") }
      let(:hg_save_path) { save_path.sub(/\.bundle$/, '.hg.bundle') }

      after do
        FileUtils.rm_rf(save_path)
        FileUtils.rm_rf(hg_save_path)
        repository.remove
      end

      it 'saves bundles to disk' do
        repository.bundle_to_disk(save_path)

        git_success = system(
          Gitlab.config.git.bin_path, "-C", git_repo_path(repository),
          "bundle", "verify", save_path,
          [:out, :err] => '/dev/null'
        )
        expect(git_success).to be true

        expect(check_hg_bundle(hg_save_path)).to be true
      end
    end

    describe '#create_from_bundle' do
      let(:repository) { non_empty_repository }
      let(:valid_bundle_path) { File.join(Dir.tmpdir, "repo-#{SecureRandom.hex}.bundle") }
      let(:valid_hg_bundle_path) { valid_bundle_path.sub(/\.bundle$/, '.hg.bundle') }
      let(:project) { create(:hg_git_project) }
      let(:imported_repo) { project.repository.raw }

      before do
        repository.set_hgrc(admin_user, true, "group",
                            { heptapod: { 'allow-bookmarks': true } })
        expect(repository.bundle_to_disk(valid_bundle_path)).to be_truthy
      end

      after do
        FileUtils.rm_rf(valid_bundle_path)
        FileUtils.rm_rf(valid_hg_bundle_path)
        repository.remove
        imported_repo.remove
      end

      it 'creates a repo from bundle files' do
        expect(imported_repo).not_to exist
        admin_user # trigger admin creation, for force_system_user

        result = imported_repo.create_from_bundle(valid_bundle_path)

        expect(result).to be_truthy
        expect(imported_repo).to exist
        expect { imported_repo.fsck }.not_to raise_exception

        # assertions depending on the hg repo itself and additional state
        hgsha = imported_repo.hgsha_from_rev('0') # needs the repo
        expect(imported_repo.sha_from_hgsha(hgsha)).not_to be_nil # needs git-mapfile
        expect(imported_repo.get_hgrc).to eq({ "allow-bookmarks" => true,
                                               "inherit" => true })
      end

      # base Git::Repository has assertions about malicious Git bundles
      # it doesn't seem useful to repeat them here, as we are now pretty
      # sure that super was correctly called
    end

    describe '#rename' do
      let(:repository) { empty_repository }

      after do
        repository.remove # whether rename succeeded or not
      end

      it 'moves the repository' do
        new_relative_path = "rename_test/relative/path.git"
        renamed_repository = described_class.new(repository.storage, new_relative_path, nil, nil)

        repository.rename(new_relative_path)

        expect(renamed_repository.exists?).to be true
        expect(as_hgitaly_repo(renamed_repository).exists?).to be true
        expect(repository.exists?).to be false
        expect(as_hgitaly_repo(repository).exists?).to be false
      end
    end

    describe '#remove' do
      let(:repository) { empty_repository }
      let(:hgitaly_repository) { as_hgitaly_repo(repository) }

      it 'removes the repository' do
        expect(repository.exists?).to be true
        expect(hgitaly_repository.exists?).to be true

        repository.remove

        expect(repository.exists?).to be false
        expect(hgitaly_repository.exists?).to be false
      end

      context 'when the repository does not exist' do
        let(:repository) { mutable_repository }

        it 'is idempotent' do
          expect(repository.exists?).to be false

          repository.remove

          expect(repository.exists?).to be false
          expect(as_hgitaly_repo(repository).exists?).to be false
        end
      end
    end
  end

  describe "#hg_changeset_branch" do
    subject { non_empty_repository }

    after do
      subject.remove
    end

    it "returns the branch, even for obsolete changeset, or proper error" do
      hgsha = subject.hgsha_from_rev('default')
      expect(subject.hg_changeset_branch(hgsha)).to eq('default')
      expect(subject).not_to be_hg_obsolete(hgsha)

      # making changeset obsolete and checking it still works
      # for that, we need a proper draft
      subject.hg_call(['phase', '-df', hgsha], user: user, for_write: true)
      subject.hg_call(['update', hgsha])
      subject.hg_call(['amend', '-e', '-m', 'amended'], user: user, for_write: true)
      expect(subject.hg_changeset_branch(hgsha)).to eq('default')
      expect(subject).to be_hg_obsolete(hgsha)

      # unknown changeset
      expect { subject.hg_changeset_branch('0123456789' * 4) }
        .to raise_error(Gitlab::Mercurial::HgError, /Could not get/)
      # invalid call
      expect { subject.hg_changeset_branch('not-a-sha') }
        .to raise_error(Gitlab::Mercurial::HgError, /hexadecimal/)
    end
  end

  describe "#hg_changeset_branch_topic" do
    subject { non_empty_repository }

    after do
      subject.remove
    end

    it "returns correct branch and (topic or nil), even for obsolete changeset" do
      sha0 = subject.hgsha_from_rev('default')

      subject.commit_files(user,
                           branch_name: 'topic/default/antilope',
                           start_branch_name: 'branch/default',
                           message: "French spelling",
                           author_email: "jdoe1@heptapod.test",
                           author_name: "John Doe1",
                           actions: [{ action: :create,
                                       file_path: 'animal',
                                       content: "antilope\n" }])
      sha1 = subject.hgsha_from_rev('antilope')

      expect(subject.hg_changeset_branch_topic(sha0)).to eq(['default', nil])
      expect(subject.hg_changeset_branch_topic(sha1)).to eq(%w[default antilope])

      # on an obsolete changeset
      subject.hg_call(['update', sha1])
      subject.hg_call(['amend', '-e', '-m', 'amended'], user: user, for_write: true)
      expect(subject.hg_changeset_branch_topic(sha1)).to eq(%w[default antilope])

      # unknown changeset
      expect { subject.hg_changeset_branch_topic('0123456789' * 4) }
        .to raise_error(Gitlab::Mercurial::HgError, /Could not get/)
      # invalid call
      expect { subject.hg_changeset_branch_topic('garbage') }
        .to raise_error(Gitlab::Mercurial::HgError, /hexadecimal/)
    end
  end

  describe "#hg_config_item_bool?" do
    subject { empty_repository }

    after do
      subject.remove
    end

    it "returns correct values" do
      subject.set_hgrc(user, true, "group",
                       { heptapod: { 'allow-bookmarks': true } })
      expect(subject.hg_config_item_bool?('heptapod', 'allow-bookmarks')).to be(true)
      expect(subject.hg_config_item_bool?('heptapod', 'allow-multiple-heads')).to be(false)

      # Used to fail with implementation based on `hg config`
      expect(subject.hg_config_item_bool?('format', 'usestore')).to be(true)
    end
  end

  describe "#create_repository" do
    subject { mutable_repository }

    after do
      subject.remove
    end

    it "makes an empty repository" do
      expect(subject.create_repository(namespace_fs_path: "group",
                                       in_subgroup: false)
            ).to be(true)

      # that's from Gitaly
      expect(subject.commit_count("branch/default")).to eq(0)

      # that's from Mercurial and it implies that we have a valid repository
      expect(subject.hgsha_from_rev('tip')).to eq(Gitlab::Mercurial::NULL_NODE)
    end
  end

  describe "#set_hgrc" do
    subject { empty_repository }

    let(:hgrc_main) { Pathname(mutable_repository_path).join('.hg', 'hgrc') }
    let(:hgrc_managed) { Pathname(mutable_repository_path).join('.hg', 'hgrc.managed') }

    after do
      subject.remove
    end

    it "writes with inherit=true a value correctly" do
      subject.set_hgrc(user, true, "group",
                       { heptapod: { 'allow-bookmarks': true } })
      expect(strip_hgrc_lines(hgrc_main.readlines))
        .to eq(["%include ../../group/hgrc", "%include hgrc.managed"])
      expect(strip_hgrc_lines(hgrc_managed.readlines))
        .to eq(["[heptapod]", "allow-bookmarks = True"])

      expect(subject.get_hgrc).to eq({ 'inherit' => true,
                                       'allow-bookmarks' => true })
    end

    it "writes with inherit=false a value correctly" do
      subject.set_hgrc(user, false, "group",
                       { heptapod: { 'auto-publish' => "nothing" } })
      expect(strip_hgrc_lines(hgrc_main.readlines))
        .to eq(["%include hgrc.managed"])
      expect(strip_hgrc_lines(hgrc_managed.readlines))
        .to eq(["[heptapod]", "auto-publish = nothing"])

      expect(subject.get_hgrc).to eq({ 'inherit' => false,
                                       'auto-publish' => 'nothing' })
    end

    it "respects inheritance with inherit=nil yet enforces inclusion of managed file" do
      subject
      hgrc_main.write("%include ../../group/hgrc\n")
      subject.set_hgrc(user, nil, "group", nil)
      expect(strip_hgrc_lines(hgrc_main.readlines))
        .to eq(["%include ../../group/hgrc", "%include hgrc.managed"])
    end
  end

  describe "#check_fix_hgrc_inheritance" do
    subject { empty_repository }

    let(:hgrc_main) { Pathname(mutable_repository_path).join('.hg', 'hgrc') }
    let(:hgrc_managed) { Pathname(mutable_repository_path).join('.hg', 'hgrc.managed') }

    after do
      subject.remove
    end

    it "fixes the inheritance relative path for a simple change" do
      subject
      hgrc_main.write("%include ../../group/hgrc\nsome other line\n")
      subject.check_fix_hgrc_inheritance("new_group")
      expect(strip_hgrc_lines(hgrc_main.readlines))
        .to eq(["%include ../../new_group/hgrc", "some other line"])
      # we would need the repo farther from repos roots from the
      # beginning to test deep paths.
    end
  end

  describe "#hg_push" do
    subject { non_empty_repository }

    let(:target) do
      repo = described_class.new('default', 'target.git', '', 'group/project')
      repo.unit_tests_skip_hooks = true
      repo.create_repository(namespace_fs_path: "group",
                             in_subgroup: false)
      # because the HGRCPATH of the HGitaly server sets the sync hook
      # and we are doing a local push, hence in the same process, we need
      # to prevent sync to run on the target (it wouldn't have the expected
      # environment for GitLab notifications). We can't (and don't want)
      # to change the environment variables of the HGitaly server running the
      # tests, obviously.
      File.open(repo.hgrc_main_path, 'w') do |hgrcf|
        hgrcf.puts('[hooks]')
        hgrcf.puts('pretxnclose.heptapod_sync=')
      end
      repo
    end

    after do
      subject.remove
      target.remove
    end

    it "performs a push" do
      hgsha = subject.hgsha_from_rev('default')
      # using lots of option to maximize test surface, which is about
      # correct passing of parameters (actual behaviour subtleties are
      # thoroughly tested in HGitaly)
      subject.hg_push(target.hg_full_path,
                      only_branches_matching: ['branch/*'],
                      include_drafts: true
                     )

      expect(target.hgsha_from_rev('default')).to eq(hgsha)
    end
  end

  describe "#rebase" do
    subject { non_empty_repository }

    after do
      subject.remove
    end

    it "is refused on named branches" do
      subject.commit_files(user,
                           branch_name: 'branch/other',
                           start_branch_name: 'branch/default',
                           message: "Commit on other branch",
                           author_email: "jdoe1@heptapod.test",
                           author_name: "John Doe1",
                           actions: [{ action: :create,
                                       file_path: 'other',
                                       content: "other foo\n" }])

      subject.commit_files(user,
                           branch_name: 'branch/default',
                           start_branch_name: 'branch/default',
                           message: "Commit on default",
                           author_email: "jdoe1@heptapod.test",
                           author_name: "John Doe1",
                           actions: [{ action: :update,
                                       file_path: 'foo',
                                       content: "foo1\n" }])

      expect do
        subject.rebase(nil, 1,
                       branch: 'branch/other',
                       branch_sha: subject.hgsha_from_rev('other'),
                       remote_repository: subject,
                       remote_branch: 'branch/default')
      end.to raise_error(Gitlab::Mercurial::HgError)
    end
  end

  describe "archive_sha" do
    subject { non_empty_repository }

    after do
      subject.remove
    end

    it "is called for the Git sha and returns the Mercurial SHA" do
      hg_sha = subject.hgsha_from_rev('default')
      git_sha = subject.find_branch('branch/default').target

      expect(subject.archive_sha(git_sha)).to eq(hg_sha)
    end
  end

  describe "commits" do
    subject { non_empty_repository }

    after do
      subject.remove
    end

    it "are retrieved from Git and Mercurial IDs and hpd_display_id is always the Mercurial ID" do
      hg_sha = subject.hgsha_from_rev('default')
      git_sha = subject.find_branch('branch/default').target

      expect(subject.commit_id_to_display_id(git_sha)).to eq(hg_sha)
      expect(subject.commit_id_from_display_id(hg_sha)).to eq(git_sha)

      commit = Gitlab::Git::Commit.find(subject, git_sha)
      expect(commit.hpd_display_id).to eq(hg_sha)

      commit = Gitlab::Git::Commit.find(subject, hg_sha)
      expect(commit.id).to eq(git_sha)
      expect(commit.hpd_display_id).to eq(hg_sha)
    end
  end

  describe "#hg_pull" do
    # we're also using non_empty_repository to pull from as an easy setup
    subject do
      main = described_class.new(
        'default', 'test_hg_pull.git', '', 'group/project')
      main.unit_tests_skip_hooks = true
      status = main.create_repository(namespace_fs_path: "group",
                                      in_subgroup: false)
      expect(status).to be(true)

      remote = described_class.new(
        'default', 'test_hg_pull_remote.git', '', 'group/project')
      remote.unit_tests_skip_hooks = true
      status = remote.create_repository(namespace_fs_path: "group",
                                        in_subgroup: false)
      expect(status).to be(true)
      remote.commit_files(user,
                          branch_name: 'branch/default',
                          message: "Commit remote default",
                          author_email: "jdoe1@heptapod.test",
                          author_name: "John Doe1",
                          actions: [{ action: :create,
                                      file_path: 'foo',
                                      content: "foo0\n" }])
      remote.commit_files(user,
                          branch_name: 'topic/default/gnu',
                          start_branch_name: 'branch/default',
                          message: "Commit remote default:gnu",
                          author_email: "jdoe1@heptapod.test",
                          author_name: "John Doe1",
                          actions: [{ action: :create,
                                      file_path: 'gnu',
                                      content: "Gnus are big antelopes\n" }])
      [main, remote]
    end

    after do
      subject.each(&:remove)
    end

    it 'works for full pulls' do
      main, remote = subject
      expect(main.branches).to eq([]) # help for debugging of side effects
      main.hg_pull(user, remote.hg_full_path)

      expect(main.branch_count).to eq(2)
      expect(Gitlab::Git::Commit.find(main, 'branch/default').message.strip)
        .to eq("Commit remote default")
      expect(Gitlab::Git::Commit.find(main, 'topic/default/gnu').message.strip)
        .to eq("Commit remote default:gnu")
      # nothing to pull is not an error
      main.hg_pull(user, remote.hg_full_path)
    end

    it 'works for partial pulls' do
      main, remote = subject
      expect(main.branches).to eq([])
      main.hg_pull(user, remote.hg_full_path,
                   gitlab_branches: ['topic/default/gnu'])
      expect(main.branch_count).to eq(1)
      expect(Gitlab::Git::Commit.find(main, 'topic/default/gnu').message.strip)
        .to eq("Commit remote default:gnu")
    end

    it 'raises HgError in case of pull errors' do
      main, remote = subject
      expect(main.branches).to eq([])
      expect do
        # we might later on relax that very precise case (unknown remote
        # branch), but it gives us an easy error trigger for the time being
        main.hg_pull(user, remote.hg_full_path,
                     gitlab_branches: ['branch/does-not-exist'])
      end.to raise_error(Gitlab::Mercurial::HgError)
    end
  end

  describe "publish permission" do
    # we don't need to add `user` to the project, because write
    # authorization is entirely managed at higher level (Models).
    # Should that change, then the test with topic creation would fail
    subject { project.repository.raw_repository }

    after do
      subject.remove
    end

    context "when hg_project_for_perms is provided" do
      it "is enforced automatically" do
        # we're using #commit_files for an easy check, but the point
        # is that it happens without direct involvement of the method
        expect do
          # need to give the user a real id, otherwise perm check can
          # fail for the wrong reason
          user.save!
          subject.commit_files(user,
                               branch_name: 'branch/default',
                               start_branch_name: 'branch/default',
                               message: "Public commit on default",
                               author_email: "jdoe2@heptapod.test",
                               author_name: "John Doe2",
                               actions: [{ action: :create,
                                           file_path: 'foo',
                                           content: "Cannot publish that\n" }])
        end.to raise_error(Gitlab::Mercurial::HgError)
      end

      it "allows creating draft changesets" do
        expect do
          subject.commit_files(user,
                               branch_name: 'topic/default/user',
                               start_branch_name: 'branch/default',
                               message: "Topical changeset",
                               author_email: "jdoe2@heptapod.test",
                               author_name: "John Doe2",
                               actions: [{ action: :create,
                                           file_path: 'foo',
                                           content: "Can commit that\n" }])
        end.not_to raise_error
      end

      it "lets Mercurial Publisher publish changesets" do
        user.save!
        project.add_member(user, :hg_publisher)
        expect do
          subject.commit_files(user,
                               branch_name: 'branch/non-protected',
                               start_branch_name: 'branch/default',
                               message: "New changeset",
                               author_email: "jdoe1@heptapod.test",
                               author_name: "John Doe1",
                               actions: [{ action: :create,
                                           file_path: 'bar',
                                           content: "Can publish\n" }])
        end.not_to raise_error
      end

      it "lets owner publish changesets" do
        expect do
          subject.commit_files(project.creator,
                               branch_name: 'branch/default',
                               start_branch_name: 'branch/default',
                               message: "New changeset",
                               author_email: "jdoe1@heptapod.test",
                               author_name: "John Doe1",
                               actions: [{ action: :create,
                                           file_path: 'bar',
                                           content: "Can publish\n" }])
        end.not_to raise_error
      end
    end
  end

  describe "commit_files writes" do
    context "encoding is base64" do
      subject { empty_repository }

      after do
        subject.remove
      end

      it "decodes for create and update" do
        subject.commit_files(user,
                          branch_name: 'branch/default',
                          message: "Commit 0",
                          author_email: "jdoe1@heptapod.test",
                          author_name: "John Doe1",
                          actions: [{ action: :create,
                                      file_path: 'foo',
                                      encoding: 'base64',
                                      content: "Zm9v\n" }])
        subject.expire_has_local_branches_cache
        expect(subject.empty?).to be(false)

        expect(subject.blob_at('branch/default', 'foo').data).to eq('foo')
        subject.commit_files(user,
                          branch_name: 'branch/default',
                          message: "Commit 1",
                          author_email: "jdoe1@heptapod.test",
                          author_name: "John Doe1",
                          actions: [{ action: :update,
                                      file_path: 'foo',
                                      encoding: 'base64',
                                      content: "YmFy\n" }])
        expect(subject.blob_at('branch/default', 'foo').data).to eq('bar')
      end
    end
  end

  describe "#hg_env_for_write" do
    context "without a project for permissions assessment" do
      subject { mutable_repository }

      it 'bypasses permissions for a regular user' do
        env = subject.hg_env_for_write(user)
        expect(env['HEPTAPOD_USERINFO_USERNAME']).to eq(user.username)
        expect(env['REMOTE_USER']).to be(nil)
      end

      it 'bypasses permissions for user=nil with force_system_user' do
        admin_id = admin_user.id
        env = subject.hg_env_for_write(nil, force_system_user: true)
        expect(env['REMOTE_USER']).to be(nil)
        expect(env['HGUSER']).to eq("Heptapod system")
        expect(env['HEPTAPOD_USERINFO_ID']).to eq(admin_id.to_s)
      end

      it 'raises ForbiddenError for user=nil without force_system_user' do
        expect { subject.hg_env_for_write(nil) }.to raise_error(
          Gitlab::HgAccess::ForbiddenError)
      end
    end

    context "when hg_project_for_perms is provided" do
      subject { project.repository.raw_repository }

      it 'propagates publish permission for project Owner' do
        owner = project.creator
        env = subject.hg_env_for_write(owner)
        expect(env['HEPTAPOD_USERINFO_USERNAME']).to eq(owner.username)
        expect(env['REMOTE_USER']).to eq('heptapod-publish')
      end

      it 'propagates publish permission for Maintainers' do
        owner = project.creator
        env = subject.hg_env_for_write(owner)
        project.add_maintainer(user)
        expect(env['HEPTAPOD_USERINFO_USERNAME']).to eq(owner.username)
        expect(env['REMOTE_USER']).to eq('heptapod-publish')
      end

      it 'gives write permission for ordinary users' do
        # write permission check must be performed in the app layer,
        # the duty of hg_env_for_write here is to distinguish between
        # write and publish permission (there's no such check in the Git case)
        env = subject.hg_env_for_write(user)
        expect(env['HEPTAPOD_USERINFO_USERNAME']).to eq(user.username)
        expect(env['REMOTE_USER']).to eq('heptapod-write')
      end

      it 'bypasses permissions for user=nil with force_system_user' do
        admin_id = admin_user.id
        env = subject.hg_env_for_write(nil, force_system_user: true)
        expect(env['REMOTE_USER']).to be(nil)
        expect(env['HEPTAPOD_USERINFO_ID']).to eq(admin_id.to_s)
      end

      it 'raises permissions for user=nil with force_system_user and no administrator' do
        expect do
          subject.hg_env_for_write(nil, force_system_user: true)
        end.to raise_error(StandardError)
      end
    end
  end

  describe '#hg_bundle_to_disk' do
    # saving in a directory that has to be created
    let(:save_dir) { File.join(Dir.tmpdir, SecureRandom.hex) }
    let(:git_save_path) { File.join(save_dir, "hg-repo.bundle") }
    let(:hg_save_path) { File.join(save_dir, "hg-repo.hg.bundle") }
    let(:subject) { non_empty_repository }

    after do
      FileUtils.rm_rf(save_dir)
      subject.remove
    end

    it 'saves a bundle to disk' do
      subject.hg_bundle_to_disk(git_save_path)
      expect(check_hg_bundle(git_save_path)).to be true
    end
  end

  # Remove hash comments (# example), lead/trailing whitespace
  # and remove empty lines
  def strip_hgrc_lines(lines)
    lines.map { |l| l.split("#", 2).first.strip }.select { |l| !l.empty? }
  end

  # Check that a bundle is ok by running `hg log` on it,
  # returning command success as a boolean
  def check_hg_bundle(path)
    hg_no_repo("-R", hg_save_path, "log")
  end

  # Run a Mercurial command directly, from outside of any repository
  #
  # For some commands, it is important to avoid running from any
  # surrounding Mercurial repository (typically Heptapod sources in HDK and
  # CI), as it may have requirements not supported by
  # the executable used by the Rails app (e.g., persistent-nodemap)
  def hg_no_repo(*args)
    Dir.chdir("/") do
      system(
        Gitlab.config.mercurial.bin_path, *args,
        [:out] => '/dev/null'
      )
    end
  end
end
