# frozen_string_literal: true

require "spec_helper"

# rubocop:disable RSpec/FilePath
RSpec.describe "Diff access HgitalyRepository", :seed_helper do
  include Gitlab::EncodingHelper
  include RepoHelpers
  include Gitlab::Popen

  def hgitaly_repo_init
    repo = Gitlab::Mercurial::HgitalyRepository.new(
      'default', 'hgitaly-repo.git', '', 'group/project')
    repo.unit_tests_skip_hooks = true
    repo
  end

  let(:hgitaly_repo) { hgitaly_repo_init }
  let(:hg_git_repo) do
    repo = Gitlab::Mercurial::HgGitRepository.new(
      hgitaly_repo.storage,
      hgitaly_repo.relative_path,
      hgitaly_repo.gl_repository,
      hgitaly_repo.gl_project_path)
    repo.unit_tests_skip_hooks = true
    repo
  end

  before(:all) do
    repo = hgitaly_repo_init
    status = repo.create_repository(namespace_fs_path: "group", in_subgroup: false)
    expect(status).to be(true)

    user = build(:user)

    repo.commit_files(user,
                      branch_name: 'branch/default',
                      message: "Commit 0",
                      author_email: "jdoe1@heptapod.test",
                      author_name: "John Doe1",
                      actions: [{ action: :create,
                                  file_path: 'pet',
                                  content: "first_pet\nsecond_pet\n" }])
    repo.expire_has_local_branches_cache # we have no post-receive hooks
    repo.commit_files(user,
                      branch_name: 'branch/default',
                      start_branch_name: 'branch/default',
                      message: "Commit 1",
                      author_email: "jdoe1@heptapod.test",
                      author_name: "John Doe1",
                      actions: [{ action: :create,
                                  file_path: "sub/car",
                                  content: "first_car\nsecond_car\n" }])
  end

  # Same repository as :repository, seen as HgGitRepository, so that
  #
  #   1. we can use HgGitRepository as a reference for assertions on method
  #      results
  #   2. we have easy fixture preparation options

  after(:all) do
    hgitaly_repo_init.remove
  end

  shared_examples 'DiffService common integration tests' do
    let(:hg_sha1) { hgitaly_repo.hgsha_from_rev('default') }
    let(:git_sha1) { hg_git_repo.commit.sha }
    let(:hg_sha0) { hgitaly_repo.hgsha_from_rev('0') }
    let(:git_sha0) do
      hg_git_repo.hg_git_invalidate_maps!
      hg_git_repo.sha_from_hgsha(hg_sha0)
    end

    it 'repo.diff_stats tests DiffStats' do
      hg_resp = hgitaly_repo.diff_stats(hg_sha0, hg_sha1).to_a
      git_resp = hg_git_repo.diff_stats(git_sha0, git_sha1).to_a
      expect(hg_resp).to eq(git_resp)
      stats = hg_resp[0]
      expect(stats.path).to eq("sub/car")
      expect(stats.additions).to eq(2)
      expect(stats.deletions).to eq(0)
    end

    it 'repo.diff tests CommitDiff' do
      hg_resp = hgitaly_repo.diff(hg_sha0, hg_sha1).to_a[0]
      git_resp = hg_git_repo.diff(git_sha0, git_sha1).to_a[0]
      expect(hg_resp.to_json).to eq(git_resp.to_json)
    end

    it 'commit.deltas tests CommitDelta' do
      hg_commit1 = Gitlab::Git::Commit.find(hgitaly_repo, 'branch/default')
      git_commit1 = Gitlab::Git::Commit.find(hg_git_repo, 'branch/default')

      hg_resp = hg_commit1.deltas[0]
      git_resp = git_commit1.deltas[0]

      expect(hg_resp.to_json).to eq(git_resp.to_json)
    end

    # TODO: Decide if we want to test and how (probably using Request Specs)
    # RawDiff and RawPatch methods, as Rails don't parse their responses
    # anywhere and directly sent to the frontend.
    #
    # We would think about FindChangedPaths too, but for now Rails is not
    # calling any RPC for this method.

    # it 'Workhorse.send_git_patch tests RawPatch' do
    # end

    # it 'Workhorse.send_git_diff tests RawDiff' do
    # end
  end

  describe 'In fully native mode' do
    it_behaves_like 'DiffService common integration tests'
  end
end
# rubocop:enable RSpec/FilePath
