# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# frozen_string_literal: true

require 'fast_spec_helper'

RSpec.describe Gitlab::Utils::LineEndings do
  describe '.analyze' do
    context 'when input is Unix-style' do
      it 'returns the Unix end-of-line' do
        expect(described_class.analyze("abc\ndef\n")).to eq("\n")
        expect(described_class.analyze("abc\ndef")).to eq("\n")
      end
    end

    context 'when input is Windows style' do
      it 'returns the Windows end-of-line' do
        expect(described_class.analyze("abc\r\ndef\r\n")).to eq("\r\n")
        expect(described_class.analyze("abc\r\ndef")).to eq("\r\n")
      end
    end

    context 'when input is MacOS old style' do
      it 'returns the Mac classic end-of-line' do
        expect(described_class.analyze("abc\rdef\r")).to eq("\r")
        expect(described_class.analyze("abc\rdef")).to eq("\r")
      end
    end

    context 'when input does not have any line end' do
      it 'returns nil' do
        expect(described_class.analyze("just a word")).to be_nil
      end
    end

    context 'when input is mix of Unix and Windows' do
      it 'returns :mixed' do
        expect(described_class.analyze("unix\n win\r\n")).to eq(:mixed)
        expect(described_class.analyze("unix\n win\r\n one more word")).to eq(:mixed)
        expect(described_class.analyze("win\r\n unix\n")).to eq(:mixed)
        expect(described_class.analyze("win\r\n unix\n one more word")).to eq(:mixed)
      end
    end

    context 'when input is mix of Unix and old Mac' do
      it 'returns :mixed' do
        expect(described_class.analyze("unix\n mac\r")).to eq(:mixed)
        expect(described_class.analyze("unix\n mac\r one more word")).to eq(:mixed)
        expect(described_class.analyze("mac\r unix\n")).to eq(:mixed)
        expect(described_class.analyze("mac\r unix\n one more word")).to eq(:mixed)
      end
    end

    context 'when input is mix of Windows and old Mac' do
      it 'returns :mixed' do
        expect(described_class.analyze("win\r\n mac\r")).to eq(:mixed)
        expect(described_class.analyze("win\r\n mac\r one more word")).to eq(:mixed)
        expect(described_class.analyze("mac\r win\r\n")).to eq(:mixed)
        expect(described_class.analyze("mac\r win\r\n one more word")).to eq(:mixed)
      end
    end

    context 'when input is mix of everything' do
      it 'returns :mixed' do
        expect(described_class.analyze("win\r\n mac\r unix\n addendum")).to eq(:mixed)
      end
    end
  end

  describe '.replace_as_in' do
    context 'when reference content is consistent' do
      it 'performs the replacement' do
        expect(described_class.replace_as_in("a\r\n b", "reference\n content\n"))
          .to eq("a\n b")
      end
    end

    context 'when reference content is inconsistent' do
      it 'raises by default' do
        expect { described_class.replace_as_in("a\nb", "incon\n sis\r\n tent\r") }
          .to raise_error(Gitlab::Utils::LineEndings::InconsistentLineEndings)
      end

      it 'uses solve_inconsistency_as if given' do
        expect(described_class.replace_as_in("a\r\n bc\r\n d", "incon\n sis\r\n tent\r",
                                             solve_inconsistency_as: "\n"))
          .to eq("a\n bc\n d")
      end
    end
  end
end
