# frozen_string_literal: true

require 'spec_helper'

RSpec.describe Repositories::DestroyService do
  let_it_be(:user) { create(:user) }

  let!(:project) { create(:project, :repository, namespace: user.namespace) }
  let(:repository) { project.repository }
  let(:path) { repository.disk_path }

  subject { described_class.new(repository).execute }

  it 'removes the repository' do
    expect(project.gitlab_shell.repository_exists?(project.repository_storage, path + '.git')).to be_truthy

    subject

    # Because the removal happens inside a run_after_commit callback we need to
    # trigger the callback
    project.touch

    expect(project.gitlab_shell.repository_exists?(project.repository_storage, path + '.git')).to be_falsey
  end

  context 'on Mercurial project' do
    let!(:project) { create(:native_hg_project, :hg_repository, namespace: user.namespace) }

    let(:storage) { project.repository_storage }
    # we need to go low level to check both repos
    let(:git_repo) { Gitlab::Git::Repository.new(storage, path + '.git', nil, nil) }
    let(:hg_repo_class) { Gitlab::Mercurial::HgitalyRepository }
    let(:hg_repo) { hg_repo_class.new(storage, path + '.git', nil, nil) }
    let(:repos) { [git_repo, hg_repo] }

    context 'when native but still with Git conversion' do
      it 'removes the repository' do
        stub_feature_flags(hg_without_git: false)
        repos.each { |repo| expect(repo.exists?).to be_truthy }

        subject

        # See callback comment in upstream (pure Git) example
        project.touch

        repos.each do |repo|
          expect(repo.exists?).to be_falsey
        end
      end
    end

    context 'when native without Git conversion' do
      let(:repos) { [hg_repo] }

      it 'removes the repository' do
        stub_feature_flags(hg_without_git: true)
        repos.each { |repo| expect(repo.exists?).to be_truthy }

        subject

        # See callback comment in upstream (pure Git) example
        project.touch

        repos.each do |repo|
          expect(repo.exists?).to be_falsey
        end
      end
    end

    context 'when hg-git based' do
      let!(:project) { create(:hg_git_project, :hg_repository, namespace: user.namespace) }
      let(:hg_repo_class) { Gitlab::Mercurial::HgGitRepository }

      it 'removes the repository' do
        repos.each { |repo| expect(repo.exists?).to be_truthy }

        subject

        # See callback comment in upstream (pure Git) example
        project.touch

        repos.each do |repo|
          expect(repo.exists?).to be_falsey
        end
      end
    end
  end

  context 'on a read-only instance' do
    before do
      allow(Gitlab::Database).to receive(:read_only?).and_return(true)
    end

    it 'schedules the repository deletion' do
      expect(project.gitlab_shell.repository_exists?(project.repository_storage, path + '.git')).to be_truthy

      subject

      expect(project.gitlab_shell.repository_exists?(project.repository_storage, path + '.git')).to be_falsey
    end
  end

  it 'flushes the repository cache' do
    expect(repository).to receive(:before_delete)

    subject
  end

  it 'does not perform any action if repository path does not exist and returns success' do
    expect(repository).to receive(:disk_path).and_return('foo')
    expect(repository).not_to receive(:before_delete)

    expect(subject[:status]).to eq :success
  end

  it 'gracefully handles exception if the repository does not exist on disk' do
    expect(repository).to receive(:before_delete).and_raise(Gitlab::Git::Repository::NoRepository)
    expect(subject[:status]).to eq :success
  end

  context 'with a project wiki repository' do
    let(:project) { create(:project, :wiki_repo) }
    let(:repository) { project.wiki.repository }

    it 'schedules the repository deletion' do
      expect(project.gitlab_shell.repository_exists?(project.repository_storage, path + '.git')).to be_truthy

      subject

      # Because the removal happens inside a run_after_commit callback we need to
      # trigger the callback
      project.touch

      expect(project.gitlab_shell.repository_exists?(project.repository_storage, path + '.git')).to be_falsey
    end
  end
end
